Index: channels/chan_sip.c
===================================================================
--- channels/chan_sip.c	(revision 342600)
+++ channels/chan_sip.c	(working copy)
@@ -14391,7 +14391,15 @@
 {
 	struct sip_peer *peer = userdata;
 
+	if (peer->mwipvt) {
+		sip_pvt_lock(peer->mwipvt);
+	}
+	ao2_lock(peer);
 	sip_send_mwi_to_peer(peer, 0);
+	ao2_unlock(peer);
+	if (peer->mwipvt) {
+		sip_pvt_unlock(peer->mwipvt);
+	}
 }
 
 static void network_change_event_subscribe(void)
@@ -14750,8 +14758,10 @@
 		sched_yield();
 	}
 	if (!res) {
+		ao2_lock(peer);
 		sip_send_mwi_to_peer(peer, 0);
 		ast_devstate_changed(AST_DEVICE_UNKNOWN, "SIP/%s", peer->name);
+		ao2_unlock(peer);
 	}
 	if (res < 0) {
 		switch (res) {
@@ -24997,7 +25007,9 @@
 			ast_set_flag(&p->flags[1], SIP_PAGE2_DIALOG_ESTABLISHED);
 			transmit_response(p, "200 OK", req);
 			if (p->relatedpeer) {	/* Send first notification */
+				ao2_lock(p->relatedpeer); /* was WRLOCK */
 				sip_send_mwi_to_peer(p->relatedpeer, 0);
+				ao2_unlock(p->relatedpeer);
 			}
 		} else if (p->subscribed != CALL_COMPLETION) {
 
@@ -25761,15 +25773,13 @@
 		}
 		ast_app_inboxcount(mailbox_str->str, &newmsgs, &oldmsgs);
 	}
-	ao2_lock(peer);
 
 	if (peer->mwipvt) {
 		/* Base message on subscription */
-		p = dialog_ref(peer->mwipvt, "sip_send_mwi_to_peer: Setting dialog ptr p from peer->mwipvt-- should this be done?");
+		p = dialog_ref(peer->mwipvt, "sip_send_mwi_to_peer: Setting dialog ptr p from peer->mwipvt");
 	} else {
 		/* Build temporary dialog for this message */
 		if (!(p = sip_alloc(NULL, NULL, 0, SIP_NOTIFY, NULL))) {
-			ao2_unlock(peer);
 			return -1;
 		}
 
@@ -25783,7 +25793,6 @@
 			dialog_unlink_all(p);
 			dialog_unref(p, "unref dialog p just created via sip_alloc");
 			/* sip_destroy(p); */
-			ao2_unlock(peer);
 			return -1;
 		}
 		/* Recalculate our side, and recalculate Call ID */
@@ -25801,17 +25810,11 @@
 		sip_scheddestroy(p, DEFAULT_TRANS_TIMEOUT);
 	}
 
-	/* We have multiple threads (mwi events and monitor retransmits) working with this PVT and as we modify the sip history if that's turned on,
-	   we really need to have a lock on it */
-	sip_pvt_lock(p);
-
 	/* Send MWI */
 	ast_set_flag(&p->flags[0], SIP_OUTGOING);
 	/* the following will decrement the refcount on p as it finishes */
 	transmit_notify_with_mwi(p, newmsgs, oldmsgs, peer->vmexten);
-	sip_pvt_unlock(p);
 	dialog_unref(p, "unref dialog ptr p just before it goes out of scope at the end of sip_send_mwi_to_peer.");
-	ao2_unlock(peer);
 	return 0;
 }
 
